import pytest

from pdm.models.setup import Setup


@pytest.mark.parametrize(
    "content, result",
    [
        (
            """[metadata]
name = foo
version = 0.1.0
""",
            Setup("foo", "0.1.0"),
        ),
        (
            """[metadata]
name = foo
version = attr:foo.__version__
""",
            Setup("foo", "0.0.0"),
        ),
        (
            """[metadata]
name = foo
version = 0.1.0

[options]
python_requires = >=3.6
install_requires =
    click
    requests
[options.extras_require]
tui =
    rich
""",
            Setup("foo", "0.1.0", ["click", "requests"], {"tui": ["rich"]}, ">=3.6"),
        ),
    ],
)
def test_parse_setup_cfg(content, result, tmp_path):
    tmp_path.joinpath("setup.cfg").write_text(content)
    assert Setup.from_directory(tmp_path) == result


@pytest.mark.parametrize(
    "content,result",
    [
        (
            """from setuptools import setup

setup(name="foo", version="0.1.0")
""",
            Setup("foo", "0.1.0"),
        ),
        (
            """import setuptools

setuptools.setup(name="foo", version="0.1.0")
""",
            Setup("foo", "0.1.0"),
        ),
        (
            """from setuptools import setup

kwargs = {"name": "foo", "version": "0.1.0"}
setup(**kwargs)
""",
            Setup("foo", "0.1.0"),
        ),
        (
            """from setuptools import setup
name = 'foo'
setup(name=name, version="0.1.0")
""",
            Setup("foo", "0.1.0"),
        ),
        (
            """from setuptools import setup

setup(name="foo", version="0.1.0", install_requires=['click', 'requests'],
      python_requires='>=3.6', extras_require={'tui': ['rich']})
""",
            Setup("foo", "0.1.0", ["click", "requests"], {"tui": ["rich"]}, ">=3.6"),
        ),
        (
            """from pathlib import Path
from setuptools import setup

version = Path('__version__.py').read_text().strip()

setup(name="foo", version=version)
""",
            Setup("foo", "0.0.0"),
        ),
    ],
)
def test_parse_setup_py(content, result, tmp_path):
    tmp_path.joinpath("setup.py").write_text(content)
    assert Setup.from_directory(tmp_path) == result


def test_parse_pyproject_toml(tmp_path):
    content = """[project]
name = "foo"
version = "0.1.0"
requires-python = ">=3.6"
dependencies = ["click", "requests"]

[project.optional-dependencies]
tui = ["rich"]
"""
    tmp_path.joinpath("pyproject.toml").write_text(content)
    result = Setup("foo", "0.1.0", ["click", "requests"], {"tui": ["rich"]}, ">=3.6")
    assert Setup.from_directory(tmp_path) == result



import tarfile
import zipfile
from unittest import mock

import pytest

from pdm.cli.commands.build import Command

pytestmark = pytest.mark.usefixtures("local_finder")


def get_tarball_names(path):
    with tarfile.open(path, "r:gz") as tar:
        return tar.getnames()


def get_wheel_names(path):
    with zipfile.ZipFile(path) as zf:
        return zf.namelist()


def test_build_command(project, pdm, mocker):
    do_build = mocker.patch.object(Command, "do_build")
    # prevent the context_settings from being reset
    project.core.exit_stack.pop_all()
    pdm(["build", "--no-sdist", "-Ca=1", "--config-setting", "b=2"], obj=project)
    do_build.assert_called_with(
        project,
        sdist=False,
        wheel=True,
        dest=mock.ANY,
        clean=True,
        verbose=0,
        hooks=mock.ANY,
    )
    assert project.core.state.config_settings == {"a": "1", "b": "2"}


def test_build_global_project_forbidden(pdm):
    result = pdm(["build", "-g"])
    assert result.exit_code != 0


def test_build_single_module(fixture_project):
    project = fixture_project("demo-module")

    Command.do_build(project)
    tar_names = get_tarball_names(project.root / "dist/demo_module-0.1.0.tar.gz")
    for name in [
        "foo_module.py",
        "bar_module.py",
        "LICENSE",
        "pyproject.toml",
        "PKG-INFO",
    ]:
        assert f"demo_module-0.1.0/{name}" in tar_names

    for i in range(2):
        if i == 1:
            Command.do_build(project, sdist=False)
        zip_names = get_wheel_names(project.root / "dist/demo_module-0.1.0-py3-none-any.whl")
        for name in ["foo_module.py", "bar_module.py"]:
            assert name in zip_names

        for name in ("pyproject.toml", "LICENSE"):
            assert name not in zip_names


def test_build_single_module_with_readme(fixture_project):
    project = fixture_project("demo-module")
    project.pyproject.metadata["readme"] = "README.md"
    project.pyproject.write()
    Command.do_build(project)
    assert "demo_module-0.1.0/README.md" in get_tarball_names(project.root / "dist/demo_module-0.1.0.tar.gz")


def test_build_package(fixture_project):
    project = fixture_project("demo-package")
    Command.do_build(project)

    tar_names = get_tarball_names(project.root / "dist/my_package-0.1.0.tar.gz")
    assert "my_package-0.1.0/my_package/__init__.py" in tar_names
    assert "my_package-0.1.0/my_package/data.json" in tar_names
    assert "my_package-0.1.0/single_module.py" not in tar_names
    assert "my_package-0.1.0/data_out.json" not in tar_names

    zip_names = get_wheel_names(project.root / "dist/my_package-0.1.0-py3-none-any.whl")
    assert "my_package/__init__.py" in zip_names
    assert "my_package/data.json" in zip_names
    assert "single_module.py" not in zip_names
    assert "data_out.json" not in zip_names


def test_build_src_package(fixture_project):
    project = fixture_project("demo-src-package")
    Command.do_build(project)

    tar_names = get_tarball_names(project.root / "dist/demo_package-0.1.0.tar.gz")
    assert "demo_package-0.1.0/src/my_package/__init__.py" in tar_names
    assert "demo_package-0.1.0/src/my_package/data.json" in tar_names

    zip_names = get_wheel_names(project.root / "dist/demo_package-0.1.0-py3-none-any.whl")
    assert "my_package/__init__.py" in zip_names
    assert "my_package/data.json" in zip_names


def test_build_package_include(fixture_project):
    project = fixture_project("demo-package")
    build_config = project.pyproject.settings.setdefault("build", {})
    build_config["includes"] = [
        "my_package/",
        "single_module.py",
        "data_out.json",
    ]
    build_config["excludes"] = ["my_package/*.json"]
    project.pyproject.write()
    Command.do_build(project)

    tar_names = get_tarball_names(project.root / "dist/my_package-0.1.0.tar.gz")
    assert "my_package-0.1.0/my_package/__init__.py" in tar_names
    assert "my_package-0.1.0/my_package/data.json" not in tar_names
    assert "my_package-0.1.0/single_module.py" in tar_names
    assert "my_package-0.1.0/data_out.json" in tar_names

    zip_names = get_wheel_names(project.root / "dist/my_package-0.1.0-py3-none-any.whl")
    assert "my_package/__init__.py" in zip_names
    assert "my_package/data.json" not in zip_names
    assert "single_module.py" in zip_names
    assert "data_out.json" in zip_names


def test_build_src_package_by_include(fixture_project):
    project = fixture_project("demo-src-package")
    project.pyproject.settings.setdefault("build", {})["includes"] = ["src/my_package"]
    project.pyproject.write()
    Command.do_build(project)

    tar_names = get_tarball_names(project.root / "dist/demo_package-0.1.0.tar.gz")
    assert "demo_package-0.1.0/src/my_package/__init__.py" in tar_names
    assert "demo_package-0.1.0/src/my_package/data.json" in tar_names

    zip_names = get_wheel_names(project.root / "dist/demo_package-0.1.0-py3-none-any.whl")
    assert "my_package/__init__.py" in zip_names
    assert "my_package/data.json" in zip_names


def test_build_with_config_settings(fixture_project):
    project = fixture_project("demo-src-package")
    project.core.state.config_settings = {"--plat-name": "win_amd64"}
    Command.do_build(project)

    assert (project.root / "dist/demo_package-0.1.0-py3-none-win_amd64.whl").exists()


def test_cli_build_with_config_settings(fixture_project, pdm):
    project = fixture_project("demo-src-package")
    result = pdm(["build", "-C--plat-name=win_amd64"], obj=project)
    assert result.exit_code == 0
    assert (project.root / "dist/demo_package-0.1.0-py3-none-win_amd64.whl").exists()


@pytest.mark.usefixtures("local_finder")
def test_build_with_no_isolation(pdm, project):
    result = pdm(["build", "--no-isolation"], obj=project)
    assert result.exit_code == 1
    pdm(["add", "pdm-backend", "--no-self"], obj=project, strict=True)
    result = pdm(["build", "--no-isolation"], obj=project)
    assert result.exit_code == 0


def test_build_ignoring_pip_environment(fixture_project, monkeypatch):
    project = fixture_project("demo-module")
    monkeypatch.setenv("PIP_REQUIRE_VIRTUALENV", "1")
    Command.do_build(project)



import os
import sys
from pathlib import Path
from types import ModuleType, SimpleNamespace

import pytest

from pdm.environments.local import PythonLocalEnvironment
from pdm.utils import pdm_scheme


@pytest.fixture()
def local_env(project):
    # Ensure we construct a fresh local environment for each test
    env = PythonLocalEnvironment(project)
    return env


def test_packages_path_compat_suffix_32(local_env, tmp_path, monkeypatch):
    # Simulate interpreter identifier ending with -32, and only non-"-32" directory exists
    monkeypatch.setattr(local_env, "_interpreter", SimpleNamespace(identifier="3.11-32"))
    base = local_env.project.root / "__pypackages__"
    compat_dir = base / "3.11"
    compat_dir.mkdir(parents=True, exist_ok=True)
    # Also ensure parent exists for side effects inside packages_path
    p = local_env.packages_path
    assert p.name == "3.11"
    assert p.parent == base


def test_local_get_paths_headers_override(local_env):
    paths = local_env.get_paths(dist_name="mypkg")
    # Ensure headers path is under include/mypkg (cross-platform path check)
    from pathlib import Path as _P

    assert _P(paths["headers"]).parts[-2:] == ("include", "mypkg")
    # Sanity: scheme base is pep582
    scheme = pdm_scheme(local_env.packages_path.as_posix())
    for k in ("purelib", "platlib", "scripts", "data"):
        assert paths[k].startswith(scheme[k])


def test_pip_command_uses_existing_module(monkeypatch, project):
    # Simulate: `python -Esm pip --version` fails, but host pip is compatible
    env = PythonLocalEnvironment(project)

    class DummyCompleted:
        returncode = 1

    monkeypatch.setattr("subprocess.run", lambda *a, **k: DummyCompleted())

    # Provide a dummy pip module with a file path
    dummy_pip = ModuleType("pip")
    dummy_dir = Path(project.core.create_temp_dir(prefix="pip-test-"))
    dummy_file = dummy_dir / "__init__.py"
    dummy_file.write_text("")
    dummy_pip.__file__ = str(dummy_file)
    monkeypatch.setitem(sys.modules, "pip", dummy_pip)

    # Make it considered compatible; patch the symbol used in BaseEnvironment
    monkeypatch.setattr("pdm.environments.base.is_pip_compatible_with_python", lambda v: True)

    cmd = env.pip_command
    assert cmd[:3] == [str(env.interpreter.executable), "-Es", str(dummy_dir)]


def test_pip_command_download_fallback(monkeypatch, project):
    # Simulate: `python -Esm pip --version` fails and host pip is unavailable/incompatible
    env = PythonLocalEnvironment(project)

    class DummyCompleted:
        returncode = 1

    monkeypatch.setattr("subprocess.run", lambda *a, **k: DummyCompleted())

    # Force importing pip to fail inside BaseEnvironment so pip_location is None
    import builtins as _builtins

    _real_import = _builtins.__import__

    def _fake_import(name, globals=None, locals=None, fromlist=(), level=0):
        if name == "pip" or name.startswith("pip."):
            raise ImportError
        return _real_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr("builtins.__import__", _fake_import)

    # Also ensure compatibility check returns False on the symbol used in BaseEnvironment
    monkeypatch.setattr("pdm.environments.base.is_pip_compatible_with_python", lambda v: False)

    def fake_download(path):
        # Create the expected wheel path
        Path(path).write_text("")

    # Avoid network: stub out the download function to just create the file
    monkeypatch.setattr(type(env), "_download_pip_wheel", lambda self, p: fake_download(p))

    cmd = env.pip_command
    # Expect the -m form not used, fallback to the wheel's pip script
    assert cmd[0] == str(env.interpreter.executable)
    assert Path(cmd[1]).name == "pip"


def test_pip_command_installed(monkeypatch, project):
    # Simulate: `python -Esm pip --version` succeeds -> use it directly
    env = PythonLocalEnvironment(project)

    class DummyCompleted:
        returncode = 0

    monkeypatch.setattr("subprocess.run", lambda *a, **k: DummyCompleted())

    cmd = env.pip_command
    assert cmd[:3] == [str(env.interpreter.executable), "-Esm", "pip"]


def test_script_kind_posix(local_env):
    # On non-Windows platforms, script_kind should be posix
    if os.name != "nt":
        assert local_env.script_kind == "posix"


def test_which_python_variants(local_env):
    # Should resolve to interpreter path when asking for pythonN or python
    exe = str(local_env.interpreter.executable)
    assert local_env.which("python") == exe
    # python3 matches the major version
    assert local_env.which(f"python{local_env.interpreter.version.major}") == exe


def test_process_env_includes_scripts_first(local_env):
    env = local_env.process_env
    scripts = local_env.get_paths()["scripts"]
    path_entries = env["PATH"].split(os.pathsep)
    assert path_entries[0] == scripts



from types import SimpleNamespace
from unittest.mock import ANY

import pytest

from pdm.cli.commands import self_cmd


def mock_distribution(metadata, entry_points=()):
    entry_points = (SimpleNamespace(group=ep) for ep in entry_points)
    return SimpleNamespace(metadata=metadata, entry_points=entry_points)


DISTRIBUTIONS = {
    "foo": mock_distribution({"Name": "foo", "Version": "1.0.0", "Summary": "Foo package"}, ["pdm.plugin"]),
    "bar": mock_distribution({"Name": "bar", "Version": "2.0.0", "Summary": "Bar package"}, ["pdm"]),
    "baz": mock_distribution({"Name": "baz", "Version": "3.0.0", "Summary": "Baz package"}),
}


@pytest.fixture()
def mock_pip(mocker):
    mocked = mocker.patch("pdm.cli.commands.self_cmd.run_pip")
    return mocked


@pytest.fixture()
def mock_all_distributions(mocker):
    mocker.patch("pdm.cli.commands.self_cmd.WorkingSet", return_value=DISTRIBUTIONS)


@pytest.fixture()
def mock_latest_pdm_version(mocker):
    return mocker.patch(
        "pdm.cli.commands.self_cmd.get_latest_pdm_version_from_pypi",
    )


@pytest.mark.usefixtures("mock_all_distributions")
def test_self_list(pdm):
    result = pdm(["self", "list"])
    assert result.exit_code == 0, result.stderr
    packages = [line.split()[0] for line in result.stdout.splitlines()]
    assert packages == ["bar", "baz", "foo"]


@pytest.mark.usefixtures("mock_all_distributions")
def test_self_list_plugins(pdm):
    result = pdm(["self", "list", "--plugins"])
    assert result.exit_code == 0, result.stderr
    packages = [line.split()[0] for line in result.stdout.splitlines()]
    assert packages == ["bar", "foo"]


def test_self_add(pdm, mock_pip):
    result = pdm(["self", "add", "foo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["install", "foo"])

    result = pdm(["self", "add", "--pip-args", "--force-reinstall --upgrade", "foo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["install", "--force-reinstall", "--upgrade", "foo"])


def test_self_remove(pdm, mock_pip, mocker, monkeypatch):
    from rich import get_console

    console = get_console()

    def _mock_resolve(packages):
        return ["demo", "pytz"] if "demo" in packages else packages

    mocker.patch.object(
        self_cmd.RemoveCommand,
        "_resolve_dependencies_to_remove",
        side_effect=_mock_resolve,
    )
    mocker.patch.object(console, "is_interactive", True)

    result = pdm(["self", "remove", "foo"])
    assert result.exit_code != 0

    result = pdm(["self", "remove", "-y", "demo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["uninstall", "-y", "demo", "pytz"])

    with monkeypatch.context() as m:
        m.setenv("PDM_NON_INTERACTIVE", "1")
        result = pdm(["self", "remove", "demo"])
        assert result.exit_code == 0, result.stderr
        mock_pip.assert_called_with(ANY, ["uninstall", "-y", "demo", "pytz"])

    result = pdm(["-n", "self", "remove", "demo"])
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, ["uninstall", "-y", "demo", "pytz"])


@pytest.mark.parametrize(
    "args,expected",
    [
        (["self", "update"], ["install", "--upgrade", "--upgrade-strategy", "eager", "pdm[locked]==99.0.0"]),
        (["self", "update", "--pre"], ["install", "--upgrade", "--upgrade-strategy", "eager", "pdm[locked]==99.0.1b1"]),
        (
            ["self", "update", "--head"],
            ["install", "--upgrade", "--upgrade-strategy", "eager", f"pdm[locked] @ git+{self_cmd.PDM_REPO}@main"],
        ),
    ],
)
def test_self_update(pdm, mock_pip, mock_latest_pdm_version, args, expected):
    def mocked_latest_version(project, pre):
        return "99.0.1b1" if pre else "99.0.0"

    mock_latest_pdm_version.side_effect = mocked_latest_version

    result = pdm(args)
    assert result.exit_code == 0, result.stderr
    mock_pip.assert_called_with(ANY, expected)


def test_self_update_already_latest(pdm, mock_pip, mock_latest_pdm_version):
    mock_latest_pdm_version.return_value = "0.0.0"

    result = pdm(["self", "update"])
    assert result.exit_code == 0, result.stderr
    assert "Already up-to-date" in result.stdout
    mock_pip.assert_not_called()



import platform
import sys
from pathlib import Path

import pytest
from pbs_installer import PythonVersion

from pdm.models.python import PythonInfo
from pdm.utils import parse_version


@pytest.fixture
def mock_install(mocker):
    if (arch := platform.machine().lower()) not in ("arm64", "aarch64", "amd64", "x86_64"):
        pytest.skip(f"Skipped on non-standard platform: {arch}")

    def install_file(
        filename,
        destination,
        original_filename=None,
        build_dir=False,
    ) -> None:
        if sys.platform == "win32":
            Path(destination, "python.exe").touch()
        else:
            Path(destination, "bin").mkdir(parents=True, exist_ok=True)
            Path(destination, "bin", "python3").touch()

    def get_download_link(version, *args, **kwargs):
        return f"cpython@{version}", None

    def get_version(self):
        name = self.executable.parent.name if sys.platform == "win32" else self.executable.parent.parent.name
        if "@" not in name:
            return parse_version(platform.python_version())
        return parse_version(name.split("@", 1)[1].rstrip("t"))

    def get_freethreaded(self):
        name = self.executable.parent.name if sys.platform == "win32" else self.executable.parent.parent.name
        return name.endswith("t")

    @property
    def interpreter(self):
        return self.executable

    @property
    def implementation(self):
        name = self.executable.parent.name if sys.platform == "win32" else self.executable.parent.parent.name
        if "@" not in name:
            return "cpython"
        return name.split("@", 1)[0]

    mocker.patch("pbs_installer.download", return_value="python-3.10.8.tar.gz")
    mocker.patch("pbs_installer.get_download_link", side_effect=get_download_link)
    installer = mocker.patch("pbs_installer.install_file", side_effect=install_file)
    mocker.patch("findpython.python.PythonVersion.implementation", implementation)
    mocker.patch("findpython.python.PythonVersion._get_version", get_version)
    mocker.patch("findpython.python.PythonVersion._get_freethreaded", get_freethreaded)
    mocker.patch("findpython.python.PythonVersion.interpreter", interpreter)
    mocker.patch("findpython.python.PythonVersion.architecture", mocker.PropertyMock(return_value="64bit"))
    return installer


def test_install_python(project, pdm, mock_install):
    root = Path(project.config["python.install_root"])

    pdm(["py", "install", "cpython@3.10.8", "-v"], obj=project, strict=True)
    mock_install.assert_called_once()
    assert (root / "cpython@3.10.8").exists()

    result = pdm(["py", "list"], obj=project, strict=True)
    assert result.stdout.splitlines()[0].startswith("cpython@3.10.8")

    result = pdm(["py", "remove", "3.11.1"], obj=project)
    assert result.exit_code != 0
    pdm(["py", "remove", "cpython@3.10.8"], obj=project, strict=True)
    assert not (root / "cpython@3.10.8").exists()

    result = pdm(["py", "install", "--list"], obj=project, strict=True)
    assert len(result.stdout.splitlines()) > 0


def test_install_python_best_match(project, pdm, mock_install, mocker):
    root = Path(project.config["python.install_root"])
    mock_best_match = mocker.patch(
        "pdm.project.core.Project.get_best_matching_cpython_version", return_value=PythonVersion("cpython", 3, 10, 8)
    )

    pdm(["py", "install"], obj=project, strict=True)
    mock_best_match.assert_called_once()
    mock_install.assert_called_once()
    assert (root / "cpython@3.10.8").exists()


def test_install_python_min_match(project, pdm, mock_install, mocker):
    root = Path(project.config["python.install_root"])
    mock_best_match = mocker.patch(
        "pdm.project.core.Project.get_best_matching_cpython_version", return_value=PythonVersion("cpython", 3, 10, 7)
    )

    pdm(["py", "install", "--min"], obj=project, strict=True)
    mock_best_match.assert_called_once_with(True)
    mock_install.assert_called_once()
    assert (root / "cpython@3.10.7").exists()


def test_use_auto_install_missing(project, pdm, mock_install, mocker):
    root = Path(project.config["python.install_root"])
    mock_find_interpreters = mocker.patch("pdm.project.Project.find_interpreters", return_value=[])
    mock_best_match = mocker.patch("pdm.project.core.Project.get_best_matching_cpython_version")

    pdm(["use", "3.10.8"], obj=project, strict=True)
    mock_install.assert_called_once()
    mock_find_interpreters.assert_called_once()
    mock_best_match.assert_not_called()
    assert (root / "cpython@3.10.8").exists()


def test_use_auto_install_pick_latest(project, pdm, mock_install, mocker):
    root = Path(project.config["python.install_root"])
    mock_find_interpreters = mocker.patch("pdm.project.Project.find_interpreters", return_value=[])
    mock_best_match = mocker.patch(
        "pdm.project.core.Project.get_best_matching_cpython_version", return_value=PythonVersion("cpython", 3, 10, 8)
    )

    pdm(["use", "-v"], obj=project, strict=True)
    mock_install.assert_called_once()
    mock_find_interpreters.assert_called_once()
    mock_best_match.assert_called_once()
    assert len(list(root.iterdir())) == 1


def test_use_no_auto_install(project, pdm, mocker):
    installer = mocker.patch("pbs_installer.install_file")
    mock_best_match = mocker.patch("pdm.project.core.Project.get_best_matching_cpython_version")

    pdm(["use", "-f"], obj=project, strict=True)
    installer.assert_not_called()
    mock_best_match.assert_not_called()


def test_use_auto_install_strategy_max(project, pdm, mock_install, mocker):
    root = Path(project.config["python.install_root"])
    mock_find_interpreters = mocker.patch("pdm.project.Project.find_interpreters")
    mock_best_match = mocker.patch(
        "pdm.project.core.Project.get_best_matching_cpython_version", return_value=PythonVersion("cpython", 3, 10, 8)
    )

    pdm(["use", "--auto-install-max"], obj=project, strict=True)
    mock_install.assert_called_once()
    mock_find_interpreters.assert_not_called()
    mock_best_match.assert_called_once()
    assert len(list(root.iterdir())) == 1


def test_use_auto_install_strategy_min(project, pdm, mock_install, mocker):
    root = Path(project.config["python.install_root"])
    mock_find_interpreters = mocker.patch("pdm.project.Project.find_interpreters")
    mock_best_match = mocker.patch(
        "pdm.project.core.Project.get_best_matching_cpython_version", return_value=PythonVersion("cpython", 3, 10, 7)
    )

    pdm(["use", "--auto-install-min"], obj=project, strict=True)
    mock_install.assert_called_once()
    mock_find_interpreters.assert_not_called()
    mock_best_match.assert_called_once_with(True)
    assert len(list(root.iterdir())) == 1


def test_link_python(project, pdm):
    root = Path(project.config["python.install_root"])
    pdm(["python", "link", sys.executable], obj=project, strict=True)
    python_info = PythonInfo.from_path(sys.executable)
    link_name = f"{python_info.implementation}@{python_info.identifier}"
    assert (root / link_name).resolve() == Path(sys.prefix).resolve()

    pdm(["python", "remove", link_name], obj=project, strict=True)
    assert not (root / link_name).exists()

    pdm(["python", "link", sys.executable, "--name", "foo"], obj=project, strict=True)
    assert (root / "foo").resolve() == Path(sys.prefix).resolve()

    pdm(["python", "remove", "foo"], obj=project, strict=True)
    assert not (root / "foo").exists()


def test_link_python_invalid_interpreter(project, pdm):
    result = pdm(["python", "link", "/path/to/invalid/python"], obj=project)
    assert result.exit_code != 0
    assert "Invalid Python interpreter" in result.stderr

    root = Path(project.config["python.install_root"])
    root.mkdir(parents=True, exist_ok=True)
    root.joinpath("foo").touch()
    result = pdm(["python", "link", sys.executable, "--name", "foo"], obj=project)
    assert result.exit_code != 0
    assert "Link foo already exists" in result.stderr


def test_find_python(project, pdm, mock_install):
    pdm(["py", "install", "3.10.8"], obj=project, strict=True)
    pdm(["py", "install", "3.13.0"], obj=project, strict=True)
    pdm(["py", "install", "3.13.0t"], obj=project, strict=True)
    result = pdm(["py", "find", "3.10.8"], obj=project, strict=True)
    assert "3.10.8" in result.stdout

    result = pdm(["py", "find", "3.10.8", "--managed"], obj=project, strict=True)
    assert "3.10.8" in result.stdout

    result = pdm(["py", "find", "3.10.9"], obj=project)
    assert result.exit_code != 0

    result = pdm(["py", "find", "3.13", "--managed"], obj=project)
    assert "3.13.0t" not in result.stdout
    assert "3.13.0" in result.stdout

    result = pdm(["py", "find", "3.13t", "--managed"], obj=project)
    assert "3.13.0t" in result.stdout



import sys
from textwrap import dedent
from unittest.mock import ANY

import pytest

from pdm.compat import tomllib
from pdm.models.python import PythonInfo
from pdm.utils import cd

PYTHON_VERSION = f"{sys.version_info[0]}.{sys.version_info[1]}"


@pytest.fixture(autouse=True)
def enable_interactive_mode(mocker):
    from rich import get_console

    console = get_console()
    mocker.patch.object(console, "is_interactive", True)


def test_init_validate_python_requires(project_no_init, pdm):
    result = pdm(["init"], input="\n\n\n\n\n\n\n3.7\n", obj=project_no_init)
    assert result.exit_code != 0
    assert "InvalidSpecifier" in result.stderr


def test_init_command(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    pdm(["init"], input="\ntest-project\n\n\n\n\n\n\n", strict=True, obj=project_no_init)
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_new_command(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    with cd(project_no_init.root):
        pdm(["new", "myproject"], input="\ntest-project\n\n\n\n\n\n\n", strict=True)
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("myproject/pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_command_library(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    result = pdm(
        ["init"],
        input="\ntest-project\n\ny\nTest Project\n1\n\n\n\n\n",
        obj=project_no_init,
    )
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Test Project",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f">={python_version}",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "build-system": {"build-backend": "setuptools.build_meta", "requires": ["setuptools>=61"]},
        "tool": {"pdm": {"distribution": True}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


@pytest.mark.parametrize(
    "backend_choice,merged_backend",
    [
        (0, {"build-backend": "pdm.backend", "requires": ["pdm-backend", "example"]}),
        (1, {"build-backend": "setuptools.build_meta", "requires": ["setuptools>=61"]}),
    ],
)
def test_init_template_build_system(tmp_path, project_no_init, pdm, mocker, backend_choice, merged_backend):
    template_with_backend = tmp_path / "backend-template"
    template_with_backend.mkdir()
    (template_with_backend / "pyproject.toml").write_text(
        dedent(
            """
            [project]
            dynamic = ["version"]
            name = "backend-template"

            [build-system]
            requires = ["pdm-backend", "example"]
            build-backend = "pdm.backend"
            """
        )
    )

    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    result = pdm(
        ["init", str(template_with_backend)],
        input=f"\ntest-project\n\ny\nTest Project\n{backend_choice}\n\n\n\n\n",
        obj=project_no_init,
    )
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Test Project",
            "license": {"text": "MIT"},
            "name": "test-project",
            "requires-python": f">={python_version}",
            "dynamic": ["version"],
        },
        "build-system": merged_backend,
        "tool": {"pdm": {"distribution": True}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_non_interactive(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    result = pdm(["init", "-n"], obj=project_no_init)
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    do_use.assert_called_once_with(
        project_no_init,
        ANY,
        first=True,
        ignore_remembered=True,
        ignore_requires_python=True,
        save=False,
        hooks=ANY,
    )
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": project_no_init.root.name,
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_auto_create_venv(project_no_init, pdm, mocker):
    mocker.patch("pdm.models.python.PythonInfo.get_venv", return_value=None)
    project_no_init.project_config["python.use_venv"] = True
    result = pdm(["init"], input="\ntest-project\n\ny\nTest Project\n1\n\n\n\n\n", obj=project_no_init)
    assert result.exit_code == 0
    assert project_no_init.python.executable.parent.parent == project_no_init.root / ".venv"
    assert ".pdm-python" in (project_no_init.root / ".gitignore").read_text()


def test_init_auto_create_venv_specify_python(project_no_init, pdm, mocker):
    mocker.patch("pdm.models.python.PythonInfo.get_venv", return_value=None)
    project_no_init.project_config["python.use_venv"] = True
    result = pdm(
        ["init", f"--python={PYTHON_VERSION}"],
        input="\n\n\n\n\n\n\n",
        obj=project_no_init,
    )
    assert result.exit_code == 0
    assert project_no_init.python.executable.parent.parent == project_no_init.root / ".venv"


def test_init_with_backend_default_library(project_no_init, pdm):
    pdm(["init", "--backend", "flit-core"], input="\n\n\n\n\n\n\n\n\n", obj=project_no_init)
    assert project_no_init.backend.__class__.__name__ == "FlitBackend"


def test_init_with_backend_default_library_non_interactive(project_no_init, pdm):
    pdm(["init", "-n", "--backend", "flit-core"], obj=project_no_init)
    assert project_no_init.backend.__class__.__name__ == "FlitBackend"


def test_init_with_license_non_interactive(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    expected_license = "Proprietary"
    result = pdm(["init", "-n", "--license", expected_license], obj=project_no_init)
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    do_use.assert_called_once_with(
        project_no_init,
        ANY,
        first=True,
        ignore_remembered=True,
        ignore_requires_python=True,
        save=False,
        hooks=ANY,
    )
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": f"{expected_license}"},
            "name": project_no_init.root.name,
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": "0.1.0",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data


def test_init_with_project_version_non_interactive(project_no_init, pdm, mocker):
    mocker.patch(
        "pdm.cli.commands.init.get_user_email_from_git",
        return_value=("Testing", "me@example.org"),
    )
    do_use = mocker.patch("pdm.cli.commands.use.Command.do_use", return_value=PythonInfo.from_path(sys.executable))
    expected_project_version = "2.0.42"
    result = pdm(["init", "-n", "--project-version", expected_project_version], obj=project_no_init)
    assert result.exit_code == 0
    python_version = f"{project_no_init.python.major}.{project_no_init.python.minor}"
    do_use.assert_called_once_with(
        project_no_init,
        ANY,
        first=True,
        ignore_remembered=True,
        ignore_requires_python=True,
        save=False,
        hooks=ANY,
    )
    data = {
        "project": {
            "authors": [{"email": "me@example.org", "name": "Testing"}],
            "dependencies": [],
            "description": "Default template for PDM package",
            "license": {"text": "MIT"},
            "name": project_no_init.root.name,
            "requires-python": f"=={python_version}.*",
            "readme": "README.md",
            "version": f"{expected_project_version}",
        },
        "tool": {"pdm": {"distribution": False}},
    }

    with open(project_no_init.root.joinpath("pyproject.toml"), "rb") as fp:
        assert tomllib.load(fp) == data



import os
import platform
import re
import shutil
import sys
from unittest.mock import ANY

import pytest
import shellingham

from pdm.cli.commands.venv import backends
from pdm.cli.commands.venv.utils import get_venv_prefix


@pytest.fixture(params=[True, False])
def with_pip(request):
    return request.param


@pytest.fixture()
def fake_create(monkeypatch):
    def fake_create(self, location, *args, prompt=None):
        bin_dir = "Scripts" if sys.platform == "win32" else "bin"
        suffix = ".exe" if sys.platform == "win32" else ""
        (location / bin_dir).mkdir(parents=True)
        (location / bin_dir / f"python{suffix}").touch()

    monkeypatch.setattr(backends.VirtualenvBackend, "perform_create", fake_create)
    monkeypatch.setattr(backends.VenvBackend, "perform_create", fake_create)
    monkeypatch.setattr(backends.CondaBackend, "perform_create", fake_create)


@pytest.mark.usefixtures("fake_create")
def test_venv_create(pdm, project):
    project._saved_python = None
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    assert os.path.exists(venv_path)
    assert not project._saved_python


@pytest.mark.usefixtures("fake_create")
def test_venv_create_in_project(pdm, project):
    project.project_config["venv.in_project"] = True
    pdm(["venv", "create"], obj=project, strict=True)
    venv_path = project.root / ".venv"
    assert venv_path.exists()
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 1
    assert "is not empty" in result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_create_other_location(pdm, project):
    pdm(["venv", "-p", project.root.as_posix(), "create"], strict=True)
    venv_path = project.root / ".venv"
    assert venv_path.exists()
    result = pdm(["venv", "-p", project.root.as_posix(), "create"])
    assert result.exit_code == 1
    assert "is not empty" in result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_show_path(pdm, project):
    project.project_config["venv.in_project"] = True
    pdm(["venv", "create"], obj=project, strict=True)
    pdm(["venv", "create", "--name", "test"], obj=project, strict=True)
    result = pdm(["venv", "--path", "in-project"], obj=project, strict=True)
    assert result.output.strip() == str(project.root / ".venv")
    result = pdm(["venv", "--path", "test"], obj=project)
    assert result.exit_code == 0
    result = pdm(["venv", "--path", "foo"], obj=project)
    assert result.exit_code == 1


@pytest.mark.usefixtures("fake_create")
def test_venv_list(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)

    result = pdm(["venv", "list"], obj=project)
    assert result.exit_code == 0, result.stderr
    assert venv_path in result.output


@pytest.mark.usefixtures("fake_create")
def test_venv_remove(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    result = pdm(["venv", "remove", "non-exist"], obj=project)
    assert result.exit_code != 0

    result = pdm(["venv", "remove", "-y", key], obj=project)
    assert result.exit_code == 0, result.stderr

    assert not os.path.exists(venv_path)


@pytest.mark.usefixtures("fake_create")
def test_venv_recreate(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr

    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code != 0

    result = pdm(["venv", "create", "-f"], obj=project)
    assert result.exit_code == 0, result.stderr


@pytest.mark.usefixtures("venv_backends")
def test_venv_activate(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    mocker.patch("shellingham.detect_shell", return_value=("bash", None))
    result = pdm(["venv", "activate", key], obj=project)
    assert result.exit_code == 0, result.stderr
    backend = project.config["venv.backend"]

    if backend == "conda":
        assert result.output.startswith("conda activate")
    else:
        assert result.output.strip("'\"\n").endswith("activate")
        if platform.system() == "Windows":
            assert not result.output.startswith("source")
            assert not result.output.startswith("'")
        else:
            assert result.output.startswith("source")


@pytest.mark.usefixtures("venv_backends")
@pytest.mark.skipif(platform.system() == "Windows", reason="UNIX only")
def test_venv_activate_tcsh(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    mocker.patch("shellingham.detect_shell", return_value=("tcsh", None))
    result = pdm(["venv", "activate", key], obj=project)
    assert result.output.startswith("source") and result.output.strip("'\"\n").endswith("activate.csh")


@pytest.mark.usefixtures("venv_backends")
def test_venv_activate_custom_prompt(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    creator = mocker.patch("pdm.cli.commands.venv.backends.Backend.create")
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    creator.assert_called_once_with(
        None, [], False, False, prompt=project.project_config["venv.prompt"], with_pip=False
    )


def test_venv_activate_project_without_python(pdm, project):
    project._saved_python = None
    result = pdm(["venv", "activate"], obj=project)
    assert result.exit_code != 0
    assert "The project doesn't have a saved python.path" in result.stderr


@pytest.mark.usefixtures("fake_create")
def test_venv_activate_error(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project, strict=True)

    result = pdm(["venv", "activate", "foo"], obj=project)
    assert result.exit_code != 0
    assert "No virtualenv with key" in result.stderr

    project._saved_python = os.path.abspath("fake/bin/python")
    result = pdm(["venv", "activate"], obj=project)
    assert result.exit_code != 0, result.output + result.stderr
    assert "Can't activate a non-venv Python" in result.stderr


@pytest.mark.usefixtures("venv_backends")
def test_venv_activate_no_shell(pdm, mocker, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    key = os.path.basename(venv_path)[len(get_venv_prefix(project)) :]

    mocker.patch("shellingham.detect_shell", side_effect=shellingham.ShellDetectionFailure())
    result = pdm(["venv", "activate", key], obj=project)
    assert result.exit_code == 0, result.stderr
    backend = project.config["venv.backend"]

    if backend == "conda":
        assert result.output.startswith("conda activate")
    else:
        assert result.output.strip("'\"\n").endswith("activate")
        if platform.system() == "Windows":
            assert not result.output.startswith("source")
            assert not result.output.startswith("'")
        else:
            assert result.output.startswith("source")


@pytest.mark.usefixtures("fake_create")
@pytest.mark.parametrize("keep_pypackages", [True, False])
def test_venv_auto_create(pdm, mocker, project, keep_pypackages):
    creator = mocker.patch("pdm.cli.commands.venv.backends.Backend.create")
    project._saved_python = None
    if keep_pypackages:
        project.root.joinpath("__pypackages__").mkdir(exist_ok=True)
    else:
        shutil.rmtree(project.root / "__pypackages__", ignore_errors=True)
    project.project_config["python.use_venv"] = True
    pdm(["install", "--no-self"], obj=project)
    if keep_pypackages:
        creator.assert_not_called()
    else:
        creator.assert_called_once()


@pytest.mark.usefixtures("fake_create")
def test_venv_purge(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "purge"], obj=project)
    assert result.exit_code == 0, result.stderr

    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    result = pdm(["venv", "purge"], input="y", obj=project)
    assert result.exit_code == 0, result.stderr
    assert not os.path.exists(venv_path)


@pytest.mark.usefixtures("fake_create")
def test_venv_purge_force(pdm, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    result = pdm(["venv", "purge", "-f"], obj=project)
    assert result.exit_code == 0, result.stderr
    assert not os.path.exists(venv_path)


user_options = [("none", True), ("0", False), ("all", False)]


@pytest.mark.usefixtures("venv_backends")
@pytest.mark.parametrize("user_choices, is_path_exists", user_options)
def test_venv_purge_interactive(pdm, user_choices, is_path_exists, project):
    project.project_config["venv.in_project"] = False
    result = pdm(["venv", "create"], obj=project)
    assert result.exit_code == 0, result.stderr
    venv_path = re.match(r"Virtualenv (.+) is created successfully", result.output).group(1)
    result = pdm(["venv", "purge", "-i"], input=user_choices, obj=project)
    assert result.exit_code == 0, result.stderr
    assert os.path.exists(venv_path) == is_path_exists


def test_virtualenv_backend_create(project, mocker, with_pip):
    backend = backends.VirtualenvBackend(project, None)
    assert backend.ident
    mock_call = mocker.patch("subprocess.check_call")
    location = backend.create(with_pip=with_pip)
    pip_args = [] if with_pip else ["--no-pip", "--no-setuptools", "--no-wheel"]
    mock_call.assert_called_once_with(
        [
            sys.executable,
            "-m",
            "virtualenv",
            str(location),
            "-p",
            str(backend._resolved_interpreter.executable),
            *pip_args,
        ],
        stdout=ANY,
    )


def test_venv_backend_create(project, mocker, with_pip):
    backend = backends.VenvBackend(project, None)
    assert backend.ident
    mock_call = mocker.patch("subprocess.check_call")
    location = backend.create(with_pip=with_pip)
    pip_args = [] if with_pip else ["--without-pip"]
    mock_call.assert_called_once_with(
        [
            str(backend._resolved_interpreter.executable),
            "-m",
            "venv",
            str(location),
            *pip_args,
        ],
        stdout=ANY,
    )


def test_conda_backend_create(project, mocker, with_pip):
    assert project.python
    backend = backends.CondaBackend(project, "3.9")
    assert backend.ident == "3.9"
    mock_call = mocker.patch("subprocess.check_call")
    location = backend.create(with_pip=with_pip)
    pip_args = ["pip"] if with_pip else []
    mock_call.assert_called_once_with(
        [
            "conda",
            "create",
            "--yes",
            "--prefix",
            str(location),
            "python=3.9",
            *pip_args,
        ],
        stdout=ANY,
    )

    backend = backends.CondaBackend(project, None)
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"
    assert backend.ident.startswith(python_version)
    location = backend.create()
    mock_call.assert_called_with(
        [
            "conda",
            "create",
            "--yes",
            "--prefix",
            str(location),
            f"python={python_version}",
        ],
        stdout=ANY,
    )



from __future__ import annotations

import os
import sys
import venv
from pathlib import Path
from typing import TYPE_CHECKING

import pytest
from pbs_installer import PythonVersion
from pytest_httpserver import HTTPServer

from pdm.environments import PythonEnvironment
from pdm.exceptions import PdmException, ProjectError
from pdm.models.requirements import parse_requirement
from pdm.models.specifiers import PySpecSet
from pdm.models.venv import get_venv_python
from pdm.utils import cd, is_path_relative_to, parse_version

if TYPE_CHECKING:
    from pdm.project.core import Project
    from pdm.pytest import PDMCallable

PYTHON_VERSIONS = ["3.9.7", "3.10.12", "3.10.11", "3.9.0", "3.10.13", "3.9.12"]


def get_python_versions() -> list[PythonVersion]:
    python_versions = []
    for v in PYTHON_VERSIONS:
        major, minor, micro = v.split(".")
        python_versions.append(PythonVersion("cpython", int(major), int(minor), int(micro)))
    return python_versions


def test_project_python_with_pyenv_support(project, mocker, monkeypatch):
    project._saved_python = None
    project._python = None
    monkeypatch.setenv("PDM_IGNORE_SAVED_PYTHON", "1")
    mocker.patch("pdm.project.core.PYENV_ROOT", str(project.root))
    pyenv_python = project.root / "shims/python"
    if os.name == "nt":
        pyenv_python = pyenv_python.with_suffix(".bat")
    pyenv_python.parent.mkdir()
    pyenv_python.touch()
    mocker.patch(
        "findpython.python.PythonVersion._get_version",
        return_value=parse_version("3.9.0"),
    )
    mocker.patch("findpython.python.PythonVersion._get_interpreter", return_value=sys.executable)
    assert Path(project.python.path) == pyenv_python
    assert project.python.executable == Path(sys.executable)

    # Clean cache
    project._python = None

    project.project_config["python.use_pyenv"] = False
    assert Path(project.python.path) != pyenv_python


def test_project_config_items(project):
    config = project.config

    for item in ("python.use_pyenv", "pypi.url", "cache_dir"):
        assert item in config


def test_project_config_set_invalid_key(project):
    config = project.project_config

    with pytest.raises(KeyError):
        config["foo"] = "bar"


def test_project_sources_overriding_pypi(project):
    project.project_config["pypi.url"] = "https://test.pypi.org/simple"
    assert project.sources[0].url == "https://test.pypi.org/simple"

    project.pyproject.settings["source"] = [{"url": "https://example.org/simple", "name": "pypi", "verify_ssl": True}]
    assert project.sources[0].url == "https://example.org/simple"


def test_project_sources_env_var_expansion(project, monkeypatch):
    monkeypatch.setenv("PYPI_USER", "user")
    monkeypatch.setenv("PYPI_PASS", "password")
    project.project_config["pypi.url"] = "https://${PYPI_USER}:${PYPI_PASS}@test.pypi.org/simple"
    # expanded in sources
    assert project.sources[0].url == "https://user:password@test.pypi.org/simple"
    # not expanded in project config
    assert project.project_config["pypi.url"] == "https://${PYPI_USER}:${PYPI_PASS}@test.pypi.org/simple"

    project.pyproject.settings["source"] = [
        {
            "url": "https://${PYPI_USER}:${PYPI_PASS}@example.org/simple",
            "name": "pypi",
            "verify_ssl": True,
        }
    ]
    # expanded in sources
    assert project.sources[0].url == "https://user:password@example.org/simple"
    # not expanded in tool settings
    assert project.pyproject.settings["source"][0]["url"] == "https://${PYPI_USER}:${PYPI_PASS}@example.org/simple"

    project.pyproject.settings["source"] = [
        {
            "url": "https://${PYPI_USER}:${PYPI_PASS}@example2.org/simple",
            "name": "example2",
            "verify_ssl": True,
        }
    ]
    # expanded in sources
    assert project.sources[1].url == "https://user:password@example2.org/simple"
    # not expanded in tool settings
    assert project.pyproject.settings["source"][0]["url"] == "https://${PYPI_USER}:${PYPI_PASS}@example2.org/simple"


def test_global_project(tmp_path, core):
    project = core.create_project(tmp_path, True)
    assert project.is_global
    assert isinstance(project.environment, PythonEnvironment)


def test_auto_global_project(tmp_path, core):
    tmp_path.joinpath(".pdm-home").mkdir()
    (tmp_path / ".pdm-home/config.toml").write_text("[global_project]\nfallback = true\n")
    with cd(tmp_path):
        project = core.create_project(global_config=tmp_path / ".pdm-home/config.toml")
    assert project.is_global


def test_project_use_venv(project):
    project._saved_python = None
    project._python = None
    scripts = "Scripts" if os.name == "nt" else "bin"
    suffix = ".exe" if os.name == "nt" else ""
    venv.create(project.root / "venv", symlinks=True)

    project.project_config["python.use_venv"] = True
    env = project.environment
    assert env.interpreter.executable == project.root / "venv" / scripts / f"python{suffix}"
    assert not env.is_local


def test_project_packages_path(project):
    packages_path = project.environment.packages_path
    version = ".".join(map(str, sys.version_info[:2]))
    if os.name == "nt" and sys.maxsize <= 2**32:
        assert packages_path.name == version + "-32"
    else:
        assert packages_path.name == version


def test_project_auto_detect_venv(project):
    venv.create(project.root / "test_venv")

    scripts = "Scripts" if os.name == "nt" else "bin"
    suffix = ".exe" if os.name == "nt" else ""

    project.project_config["python.use_venv"] = True
    project._python = None
    project._saved_python = (project.root / "test_venv" / scripts / f"python{suffix}").as_posix()

    assert not project.environment.is_local


@pytest.mark.path
def test_ignore_saved_python(project, monkeypatch):
    project.project_config["python.use_venv"] = True
    project._python = None
    scripts = "Scripts" if os.name == "nt" else "bin"
    suffix = ".exe" if os.name == "nt" else ""
    venv.create(project.root / "venv", symlinks=True)
    monkeypatch.setenv("PDM_IGNORE_SAVED_PYTHON", "1")
    assert project.python.executable != project._saved_python
    assert project.python.executable == project.root / "venv" / scripts / f"python{suffix}"


def test_select_dependencies(project):
    project.pyproject.metadata["dependencies"] = ["requests"]
    project.pyproject.metadata["optional-dependencies"] = {
        "security": ["cryptography"],
        "venv": ["virtualenv"],
    }
    project.pyproject.dependency_groups.update(
        {
            "test": ["pytest"],
            "doc": ["mkdocs"],
            "all": [{"include-group": "test"}, {"include-group": "doc"}],
        }
    )
    assert sorted([r.key for r in project.get_dependencies()]) == ["requests"]
    assert sorted([r.key for r in project.get_dependencies("security")]) == ["cryptography"]
    assert sorted([r.key for r in project.get_dependencies("test")]) == ["pytest"]
    assert sorted([r.key for r in project.get_dependencies("all")]) == ["mkdocs", "pytest"]
    assert sorted(project.iter_groups()) == [
        "all",
        "default",
        "doc",
        "security",
        "test",
        "venv",
    ]


def test_invalid_dependency_group(project):
    project.pyproject.dependency_groups.update(
        {
            "invalid": [{"invalid-key": True}],
            "missing": [{"include-group": "missing-group"}],
            "doc": ["mkdocs"],
            "recursive": [{"include-group": "invalid"}],
        }
    )
    assert sorted([r.key for r in project.get_dependencies("doc")]) == ["mkdocs"]
    with pytest.raises(ProjectError, match="Invalid dependency group item"):
        project.get_dependencies("invalid")
    with pytest.raises(ProjectError, match="Invalid dependency group item"):
        project.get_dependencies("recursive")
    with pytest.raises(ProjectError, match="Missing group 'missing-group'"):
        project.get_dependencies("missing")


@pytest.mark.path
def test_set_non_exist_python_path(project_no_init):
    project_no_init._saved_python = "non-exist-python"
    project_no_init._python = None
    assert project_no_init.python.executable.name != "non-exist-python"


@pytest.mark.usefixtures("venv_backends")
def test_create_venv_first_time(pdm, project, local_finder):
    project.project_config.update({"venv.in_project": False})
    project._saved_python = None
    result = pdm(["install"], obj=project)
    assert result.exit_code == 0
    venv_parent = project.root / "venvs"
    venv_path = next(venv_parent.iterdir(), None)
    assert venv_path is not None

    assert Path(project._saved_python).relative_to(venv_path)


@pytest.mark.usefixtures("venv_backends", "local_finder")
@pytest.mark.parametrize("with_pip", [True, False])
def test_create_venv_in_project(pdm, project, with_pip):
    project.project_config.update({"venv.in_project": True, "venv.with_pip": with_pip})
    project._saved_python = None
    result = pdm(["install"], obj=project)
    assert result.exit_code == 0
    assert project.root.joinpath(".venv").exists()
    working_set = project.environment.get_working_set()
    assert ("pip" in working_set) is with_pip


@pytest.mark.usefixtures("venv_backends")
def test_find_interpreters_from_venv(pdm, project, local_finder):
    project.project_config.update({"venv.in_project": False})
    project._saved_python = None
    result = pdm(["install"], obj=project)
    assert result.exit_code == 0
    venv_parent = project.root / "venvs"
    venv_path = next(venv_parent.iterdir(), None)
    venv_python = get_venv_python(venv_path)

    assert any(venv_python == p.executable for p in project.find_interpreters())


@pytest.mark.usefixtures("local_finder")
def test_find_interpreters_without_duplicate_relative_paths(pdm, project):
    project._saved_python = None
    venv.create(project.root / ".venv", clear=True)
    with cd(project.root):
        bin_dir = "Scripts" if os.name == "nt" else "bin"
        suffix = ".exe" if os.name == "nt" else ""
        found = list(project.find_interpreters(f".venv/{bin_dir}/python{suffix}"))
        assert len(found) == 1


def test_iter_project_venvs(project):
    from pdm.cli.commands.venv.utils import get_venv_prefix, iter_venvs
    from pdm.models.venv import get_venv_python

    venv_parent = Path(project.config["venv.location"])
    venv_prefix = get_venv_prefix(project)
    for name in ("foo", "bar", "baz"):
        venv_python = get_venv_python(venv_parent / (venv_prefix + name))
        venv_python.parent.mkdir(parents=True)
        venv_python.touch()
    dot_venv_python = get_venv_python(project.root / ".venv")
    dot_venv_python.parent.mkdir(parents=True)
    dot_venv_python.touch()
    venv_keys = [key for key, _ in iter_venvs(project)]
    assert sorted(venv_keys) == ["bar", "baz", "foo", "in-project"]


def test_load_extra_sources(project):
    project.pyproject.settings["source"] = [
        {
            "name": "custom",
            "url": "https://custom.pypi.org/simple",
        }
    ]
    project.global_config["pypi.extra.url"] = "https://extra.pypi.org/simple"
    sources = project.sources
    assert len(sources) == 3
    assert [item.name for item in sources] == ["pypi", "custom", "extra"]

    project.global_config["pypi.ignore_stored_index"] = True
    sources = project.sources
    assert len(sources) == 1
    assert [item.name for item in sources] == ["custom"]


def test_no_index_raise_error(project):
    project.global_config["pypi.ignore_stored_index"] = True
    with pytest.raises(PdmException, match="You must specify at least one index"):
        with project.environment.get_finder():
            pass


def test_access_index_with_auth(project, httpserver: HTTPServer):
    httpserver.expect_request(
        "/simple/my-package", method="GET", headers={"Authorization": "Basic Zm9vOmJhcg=="}
    ).respond_with_data("OK")
    project.global_config.update(
        {
            "pypi.extra.url": httpserver.url_for("/simple"),
            "pypi.extra.username": "foo",
            "pypi.extra.password": "bar",
        }
    )
    session = project.environment.session
    resp = session.get(httpserver.url_for("/simple/my-package"))
    assert resp.is_success


def test_configured_source_overwriting(project):
    project.pyproject.settings["source"] = [
        {
            "name": "custom",
            "url": "https://custom.pypi.org/simple",
        }
    ]
    project.global_config["pypi.custom.url"] = "https://extra.pypi.org/simple"
    project.global_config["pypi.custom.verify_ssl"] = False
    project.project_config["pypi.custom.username"] = "foo"
    project.project_config["pypi.custom.password"] = "bar"
    sources = project.sources
    assert [source.name for source in sources] == ["pypi", "custom"]
    custom_source = sources[1]
    assert custom_source.url == "https://custom.pypi.org/simple"
    assert custom_source.verify_ssl is False
    assert custom_source.username == "foo"
    assert custom_source.password == "bar"


def test_invoke_pdm_adding_configured_args(project, pdm, mocker):
    project.pyproject.settings["options"] = {
        "install": ["--no-self", "--no-editable"],
        "add": ["--no-isolation"],
        "lock": ["--no-cross-platform"],
    }
    project.pyproject.write()
    parser = mocker.patch("argparse.ArgumentParser.parse_args")
    pdm(["add", "requests"], obj=project)
    parser.assert_called_with(["add", "--no-isolation", "requests"])
    pdm(["install", "--check"], obj=project)
    parser.assert_called_with(["install", "--no-self", "--no-editable", "--check"])
    pdm(["lock", "--lockfile", "pdm.2.lock"], obj=project)
    parser.assert_called_with(["lock", "--no-cross-platform", "--lockfile", "pdm.2.lock"])
    pdm(["-c", "/dev/null", "lock"], obj=project)
    parser.assert_called_with(["-c", "/dev/null", "lock", "--no-cross-platform"])
    pdm(["--verbose", "add", "requests"], obj=project)
    parser.assert_called_with(["--verbose", "add", "--no-isolation", "requests"])


def test_invoke_pdm_options_loaded_from_target_project(project, pdm, mocker, tmp_path):
    """Regression test for https://github.com/pdm-project/pdm/issues/3756.

    When -p/--project is used, [tool.pdm.options] must be read from the
    target project, not from the current working directory.
    """

    # Set options on the current (cwd) project — these must NOT be applied.
    project.pyproject.settings["options"] = {"install": ["--no-editable"]}
    project.pyproject.write()

    # Build a second project at a tmp location with different options.
    target_root = tmp_path / "target_project"
    target_root.mkdir()
    pyproject = target_root / "pyproject.toml"
    pyproject.write_text("[tool.pdm.options]\ninstall = ['--no-self']\n")

    handle = mocker.patch("pdm.core.Core.handle")
    # Invoke WITHOUT obj= so the real project-path resolution path is exercised.
    pdm(["install", "-p", str(target_root)])
    # Should use --no-self from target_project, NOT --no-editable from cwd project.
    args_used = handle.call_args[0][-1]
    assert args_used.no_self, "Expected --no-self in args"
    assert not args_used.no_editable, "Did not expect --no-editable in args"


@pytest.fixture()
def prepare_repository(repository, project):
    repository.add_candidate("foo", "3.0", ">=3.8,<3.13")
    repository.add_candidate("foo", "2.0", ">=3.7,<3.12")
    repository.add_candidate("foo", "1.0", ">=3.7")
    project.environment.python_requires = PySpecSet(">=3.9")
    project.add_dependencies(["foo"])


@pytest.mark.usefixtures("prepare_repository")
@pytest.mark.parametrize("is_quiet,extra_args", [(True, ("-q",)), (False, ())])
def test_quiet_mode(pdm, project, is_quiet, extra_args, recwarn):
    result = pdm(["lock", *extra_args], obj=project)

    assert result.exit_code == 0
    assert len(recwarn) > 0

    assert 'For example, "<3.13,>=3.9"' in str(recwarn[0].message)
    assert 'For example, "<3.12,>=3.9"' in str(recwarn[1].message)
    assert ("to suppress these warnings" in result.stderr) is not is_quiet
    assert project.get_locked_repository().candidates["foo"].version == "1.0"


@pytest.mark.usefixtures("prepare_repository")
@pytest.mark.parametrize("pattern,suppressed", [("foo", True), ("bar", False), ("*", True), ("f?o", True)])
def test_ignore_package_warning(pdm, project, recwarn, pattern, suppressed):
    project.pyproject.settings["ignore_package_warnings"] = [pattern]
    result = pdm(["lock"], obj=project)

    assert result.exit_code == 0
    assert (len(recwarn) == 0) is suppressed


def test_filter_sources_with_config(project):
    project.pyproject.settings["source"] = [
        {"name": "source1", "url": "https://source1.org/simple", "include_packages": ["foo", "foo-*"]},
        {
            "name": "source2",
            "url": "https://source2.org/simple",
            "include_packages": ["foo-bar", "bar*"],
            "exclude_packages": ["baz-*"],
        },
        {"name": "pypi", "url": "https://pypi.org/simple"},
    ]
    repository = project.get_repository()

    def expect_sources(requirement: str, expected: list[str]) -> None:
        sources = repository.get_filtered_sources(parse_requirement(requirement))
        assert sorted([source.name for source in sources]) == sorted(expected)

    expect_sources("foo", ["source1"])
    expect_sources("foo-baz", ["source1"])
    expect_sources("foo-bar", ["source1", "source2"])
    expect_sources("bar-extra", ["source2"])
    expect_sources("baz-extra", ["source1", "pypi"])


@pytest.mark.parametrize("use_venv", [pytest.param(True, marks=pytest.mark.network), False])
def test_find_interpreters_with_PDM_IGNORE_ACTIVE_VENV(
    pdm: PDMCallable,
    project: Project,
    monkeypatch: pytest.MonkeyPatch,
    use_venv: bool,
):
    project._saved_python = None
    project._python = None
    project.project_config["python.use_venv"] = use_venv
    venv.create(venv_path := project.root / "venv", symlinks=True)
    monkeypatch.setenv("VIRTUAL_ENV", str(venv_path))
    monkeypatch.setenv("PDM_IGNORE_ACTIVE_VENV", "1")

    venv_python = get_venv_python(venv_path)
    pythons = list(project.find_interpreters())

    assert pythons, "PDM should find interpreters with PDM_IGNORE_ACTIVE_VENV"
    # Test requires that some interpreters are available outside the venv
    assert any(venv_python != p.executable for p in project.find_interpreters())
    # No need to assert, exception raised if not found
    interpreter = project.resolve_interpreter()
    assert interpreter.executable != venv_python

    if use_venv:
        project.project_config["venv.in_project"] = True
        pdm("install", strict=True, obj=project)
        assert project._saved_python
        python = Path(project._saved_python)
        assert is_path_relative_to(python, project.root)
        assert not is_path_relative_to(python, venv_path)


@pytest.mark.parametrize(
    "var,key,settings,expected",
    [
        ("PDM_VAR", "var", {}, "from-env"),
        ("pdm_var", "var", {}, "from-env"),
        ("PDM_NOPE", "var", {"var": "from-settings"}, "from-settings"),
        ("PDM_VAR", "var", {"var": "from-settings"}, "from-env"),
        ("PDM_NOPE", "nested.var", {"nested": {"var": "from-settings"}}, "from-settings"),
        ("PDM_NOPE", "noop", {}, None),
    ],
)
def test_env_or_setting(
    project: Project,
    monkeypatch: pytest.MonkeyPatch,
    var: str,
    key: str,
    settings: dict,
    expected: str | None,
):
    monkeypatch.setenv("PDM_VAR", "from-env")
    project.pyproject.settings.update(settings)

    assert project.env_or_setting(var, key) == expected


@pytest.mark.parametrize(
    "var,setting,expected",
    [
        (None, None, []),
        ("", None, []),
        (" ", None, []),
        (None, "", []),
        (None, " ", []),
        (None, [], []),
        ("var", None, ["var"]),
        ("val1,val2", None, ["val1", "val2"]),
        ("val1, val2", None, ["val1", "val2"]),
        ("val1, , , val2", None, ["val1", "val2"]),
        (None, "val1,val2", ["val1", "val2"]),
        (None, ["val1", "val2"], ["val1", "val2"]),
        (None, [" val1", "val2 "], ["val1", "val2"]),
        (None, [" val1", "", "val2 ", " "], ["val1", "val2"]),
        (None, ["val1,val2", "val3,val4"], ["val1,val2", "val3,val4"]),
        ("val1,val2", ["val3", "val4"], ["val1", "val2"]),
    ],
)
def test_env_setting_list(
    project: Project,
    monkeypatch: pytest.MonkeyPatch,
    var: str | None,
    setting: str | list[str] | None,
    expected: list[str],
):
    if var is not None:
        monkeypatch.setenv("PDM_VAR", var)
    if setting is not None:
        project.pyproject.settings["var"] = setting

    assert project.environment._setting_list("PDM_VAR", "var") == expected


def test_project_best_match_max(project, mocker):
    expected = PythonVersion("cpython", 3, 10, 13)
    mocker.patch(
        "pdm.project.core.get_all_installable_python_versions",
        return_value=get_python_versions(),
    )
    assert project.get_best_matching_cpython_version() == expected


def test_project_best_match_min(project, mocker):
    expected = PythonVersion("cpython", 3, 9, 0)
    mocker.patch(
        "pdm.project.core.get_all_installable_python_versions",
        return_value=get_python_versions(),
    )
    assert project.get_best_matching_cpython_version(use_minimum=True) == expected


def test_default_lockfile_format(project):
    assert project.lockfile._path.name == "pdm.lock"
    project.project_config["lock.format"] = "pylock"
    project._lockfile = None
    assert project.lockfile._path.name == "pylock.toml"

    with pytest.raises(ValueError):
        project.project_config["lock.format"] = "invalid"


def test_select_lockfile_format(project, pdm, capsys):
    pdm(["lock"], obj=project, strict=True)

    assert project.lockfile._path.name == "pdm.lock"
    project.project_config["lock.format"] = "pylock"
    project._lockfile = None
    capsys.readouterr()
    assert project.lockfile._path.name == "pdm.lock"
    _, err = capsys.readouterr()
    assert "`lock.format` is set to pylock but pylock.toml is not found" in err

    with cd(project.root):
        pdm(["export", "-f", "pylock", "-o", "pylock.toml"], strict=True)
    project._lockfile = None
    assert project.lockfile._path.name == "pylock.toml"



import pytest
from resolvelib.resolvers import ResolutionImpossible

from pdm.cli.actions import resolve_candidates_from_lockfile
from pdm.exceptions import PackageWarning
from pdm.models.markers import EnvSpec
from pdm.models.requirements import parse_requirement
from pdm.models.specifiers import PySpecSet
from pdm.project.lockfile import FLAG_DIRECT_MINIMAL_VERSIONS, FLAG_INHERIT_METADATA
from pdm.resolver.reporters import LockReporter
from pdm.utils import get_requirement_from_override
from tests import FIXTURES


@pytest.fixture()
def resolve(project, repository):
    def resolve_func(
        lines,
        requires_python=None,
        allow_prereleases=None,
        strategy="all",
        tracked_names=None,
        direct_minimal_versions=False,
        inherit_metadata=False,
        platform=None,
    ):
        env_spec = project.environment.allow_all_spec
        replace_dict = {}
        if requires_python:
            replace_dict["requires_python"] = PySpecSet(requires_python)
        if platform:
            replace_dict["platform"] = platform
        env_spec = env_spec.replace(**replace_dict)
        if allow_prereleases is not None:
            project.pyproject.settings.setdefault("resolution", {})["allow-prereleases"] = allow_prereleases
        requirements = []
        for line in lines:
            if line.startswith("-e "):
                requirements.append(parse_requirement(line[3:], True))
            else:
                requirements.append(parse_requirement(line))

        ui = project.core.ui
        strategies = project.lockfile.default_strategies.copy()
        if inherit_metadata:
            strategies.add(FLAG_INHERIT_METADATA)
        if direct_minimal_versions:
            strategies.add(FLAG_DIRECT_MINIMAL_VERSIONS)

        with ui.logging("lock"):
            resolver = project.get_resolver()(
                environment=project.environment,
                requirements=requirements,
                update_strategy=strategy,
                strategies=strategies,
                target=env_spec,
                tracked_names=tracked_names,
                reporter=LockReporter(),
            )
            return resolver.resolve().candidates

    return resolve_func


def test_resolve_named_requirement(resolve):
    result = resolve(["requests"])

    assert result["requests"].version == "2.19.1"
    assert result["urllib3"].version == "1.22"
    assert result["chardet"].version == "3.0.4"
    assert result["certifi"].version == "2018.11.17"
    assert result["idna"].version == "2.7"


def test_resolve_exclude(resolve, project):
    project.pyproject.settings.setdefault("resolution", {})["excludes"] = ["urllib3"]
    result = resolve(["requests"])

    assert result["requests"].version == "2.19.1"
    assert result["chardet"].version == "3.0.4"
    assert result["certifi"].version == "2018.11.17"
    assert result["idna"].version == "2.7"
    assert "urllib3" not in result


def test_resolve_requires_python(resolve, project):
    project.environment.python_requires = PySpecSet(">=2.7")
    with pytest.warns(PackageWarning) as records:
        result = resolve(["django"])
    assert len(records) > 0
    assert result["django"].version == "1.11.8"
    assert "sqlparse" not in result

    result = resolve(["django"], ">=3.6")
    assert result["django"].version == "2.2.9"
    assert "sqlparse" in result

    result = resolve(["django; python_version>='3.7'"])
    assert result["django"].version == "2.2.9"
    assert "sqlparse" in result


def test_resolve_allow_prereleases(resolve, repository):
    repository.add_candidate("foo", "1.0.0")
    repository.add_candidate("foo", "1.1.0-alpha")
    repository.add_candidate("bar", "1.0.0-beta")

    result = resolve(["foo"])
    assert result["foo"].version == "1.0.0"

    result = resolve(["foo"], allow_prereleases=True)
    assert result["foo"].version == "1.1.0-alpha"

    result = resolve(["foo==1.1.0a0"])
    assert result["foo"].version == "1.1.0-alpha"

    result = resolve(["bar"])
    assert result["bar"].version == "1.0.0-beta"

    with pytest.raises(ResolutionImpossible):
        resolve(["bar"], allow_prereleases=False)


def test_resolve_prereleases_if_disabled_by_project(resolve):
    result = resolve(["urllib3==1.23b0"], allow_prereleases=False)
    assert result["urllib3"].version == "1.23b0"


def test_resolve_with_extras(resolve):
    result = resolve(["requests[socks]"])
    assert result["pysocks"].version == "1.5.6"
    assert result["urllib3"].version == "1.22"
    assert result["chardet"].version == "3.0.4"
    assert result["certifi"].version == "2018.11.17"
    assert result["idna"].version == "2.7"
    assert result["requests"].version == "2.19.1"


def test_resolve_with_extras_and_excludes(resolve, project):
    project.pyproject.settings.setdefault("resolution", {})["excludes"] = ["requests"]
    result = resolve(["requests[socks]"])
    assert result["pysocks"].version == "1.5.6"
    assert "requests" not in result
    assert "urllib3" not in result


@pytest.mark.parametrize(
    "requirement_line",
    [
        f"{(FIXTURES / 'artifacts/demo-0.0.1.tar.gz').as_posix()}",
        f"{(FIXTURES / 'artifacts/demo-0.0.1-py2.py3-none-any.whl').as_posix()}",
    ],
    ids=["sdist", "wheel"],
)
def test_resolve_local_artifacts(resolve, requirement_line):
    result = resolve([requirement_line], ">=3.6")
    assert result["idna"].version == "2.7"


@pytest.mark.parametrize(
    "line",
    [
        (FIXTURES / "projects/demo").as_posix(),
        "git+https://github.com/test-root/demo.git#egg=demo",
    ],
)
def test_resolve_vcs_and_local_requirements(resolve, line, is_editable, vcs):
    editable = "-e " if is_editable else ""
    result = resolve([editable + line], ">=3.6")
    assert result["idna"].version == "2.7"


def test_resolve_vcs_without_explicit_name(resolve, vcs):
    requirement = "git+https://github.com/test-root/demo.git"
    result = resolve([requirement], ">=3.6")
    assert result["idna"].version == "2.7"


def test_resolve_local_and_named_requirement(resolve, vcs):
    requirements = ["demo", "git+https://github.com/test-root/demo.git#egg=demo"]
    result = resolve(requirements, ">=3.6")
    assert result["demo"].req.is_vcs

    requirements = ["git+https://github.com/test-root/demo.git#egg=demo", "demo"]
    result = resolve(requirements, ">=3.6")
    assert result["demo"].req.is_vcs


def test_resolving_auto_avoid_conflicts(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_candidate("foo", "0.2.0")
    repository.add_dependencies("foo", "0.1.0", ["hoho<2.0"])
    repository.add_dependencies("foo", "0.2.0", ["hoho>=2.0"])
    repository.add_candidate("bar", "0.1.0")
    repository.add_dependencies("bar", "0.1.0", ["hoho~=1.1"])
    repository.add_candidate("hoho", "2.1")
    repository.add_candidate("hoho", "1.5")

    result = resolve(["foo", "bar"])
    assert result["foo"].version == "0.1.0"
    assert result["bar"].version == "0.1.0"
    assert result["hoho"].version == "1.5"


def test_resolve_conflicting_dependencies(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["hoho>=2.0"])
    repository.add_candidate("bar", "0.1.0")
    repository.add_dependencies("bar", "0.1.0", ["hoho~=1.1"])
    repository.add_candidate("hoho", "2.1")
    repository.add_candidate("hoho", "1.5")
    with pytest.raises(ResolutionImpossible):
        resolve(["foo", "bar"])


@pytest.mark.parametrize("overrides", ["2.1", ">=1.8", "==2.1"])
def test_resolve_conflicting_dependencies_with_overrides(project, resolve, repository, overrides):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["hoho>=2.0"])
    repository.add_candidate("bar", "0.1.0")
    repository.add_dependencies("bar", "0.1.0", ["hoho~=1.1"])
    repository.add_candidate("baz", "0.1.0")
    repository.add_dependencies("baz", "0.1.0", ["hoho[extra]~=1.1"])
    repository.add_candidate("hoho", "2.1")
    repository.add_candidate("hoho", "1.5")
    project.pyproject.settings["resolution"] = {"overrides": {"hoho": overrides}}
    result = resolve(["foo", "bar"])
    assert result["hoho"].version == "2.1"
    result = resolve(["foo", "baz"])
    assert result["hoho"].version == "2.1"


def test_resolve_no_available_versions(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    with pytest.raises(ResolutionImpossible):
        resolve(["foo>=0.2.0"])


def test_exclude_incompatible_requirements(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["bar; python_version < '3'"])
    result = resolve(["foo"], ">=3.6")
    assert "bar" not in result


def test_union_markers_from_different_parents(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["bar; python_version < '3'"])
    repository.add_candidate("bar", "0.1.0")
    result = resolve(["foo", "bar"], ">=3.6")
    assert not result["bar"].requires_python


def test_requirements_from_different_groups(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_candidate("foo", "0.2.0")
    requirements = ["foo", "foo<0.2.0"]
    result = resolve(requirements)
    assert result["foo"].version == "0.1.0"


def test_resolve_two_extras_from_the_same_package(resolve):
    # Case borrowed from pypa/pip#7096
    line = (FIXTURES / "projects/demo_extras").as_posix() + "[extra1,extra2]"
    result = resolve([line])
    assert "pysocks" in result
    assert "pyopenssl" in result


def test_resolve_package_with_dummy_upbound(resolve, repository):
    repository.add_candidate("foo", "0.1.0", ">=3.5,<4.0")
    result = resolve(["foo"], ">=3.5")
    assert "foo" in result


def test_resolve_dependency_with_extra_marker(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["pytz; extra=='tz' or extra=='all'"])
    result = resolve(["foo"])
    assert "pytz" not in result

    result = resolve(["foo[tz]"])
    assert "pytz" in result


def test_resolve_circular_dependencies(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["foobar"])
    repository.add_candidate("foobar", "0.2.0")
    repository.add_dependencies("foobar", "0.2.0", ["foo"])
    result = resolve(["foo"])
    assert result["foo"].version == "0.1.0"
    assert result["foobar"].version == "0.2.0"


def test_resolve_candidates_to_install(project):
    project.lockfile.set_data(
        {
            "metadata": {"strategy": ["cross_platform"]},
            "package": [
                {
                    "name": "pytest",
                    "version": "4.6.0",
                    "summary": "pytest module",
                    "dependencies": ["py>=3.0", "configparser; sys_platform=='win32'"],
                },
                {
                    "name": "configparser",
                    "version": "1.2.0",
                    "summary": "configparser module",
                    "dependencies": ["backports"],
                },
                {
                    "name": "py",
                    "version": "3.6.0",
                    "summary": "py module",
                },
                {
                    "name": "backports",
                    "version": "2.2.0",
                    "summary": "backports module",
                },
            ],
        }
    )
    reqs = [parse_requirement("pytest")]
    result = resolve_candidates_from_lockfile(project, reqs, env_spec=EnvSpec.from_spec("==3.11", "linux", "cpython"))
    assert result["pytest"].version == "4.6.0"
    assert result["py"].version == "3.6.0"
    assert "configparser" not in result
    assert "backports" not in result

    result = resolve_candidates_from_lockfile(project, reqs, env_spec=EnvSpec.from_spec("==3.11", "windows", "cpython"))
    assert result["pytest"].version == "4.6.0"
    assert result["py"].version == "3.6.0"
    assert result["configparser"].version == "1.2.0"
    assert result["backports"].version == "2.2.0"


def test_resolve_prefer_requirement_with_prereleases(resolve):
    result = resolve(["urllib3", "requests>=2.20.0b0"])
    assert result["urllib3"].version == "1.23b0"


def test_resolve_with_python_marker(resolve):
    result = resolve(["demo; python_version>='3.6'"])
    assert result["demo"].version == "0.0.1"


def test_resolve_file_req_with_prerelease(resolve, vcs):
    result = resolve(
        [
            "using-demo==0.1.0",
            "demo @ git+https://github.com/test-root/demo-prerelease.git",
        ],
        ">=3.6",
        allow_prereleases=False,
    )
    assert result["demo"].version == "0.0.2b0"


def test_resolve_extra_requirements_no_break_constraints(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["chardet; extra=='chardet'"])
    repository.add_candidate("foo", "0.2.0")
    repository.add_dependencies("foo", "0.2.0", ["chardet; extra=='chardet'"])
    result = resolve(["foo[chardet]<0.2.0"])
    assert "chardet" in result
    assert result["foo"].version == "0.1.0"


def test_resolve_extra_and_underlying_to_the_same_version(resolve, repository):
    repository.add_candidate("foo", "0.1.0")
    repository.add_dependencies("foo", "0.1.0", ["chardet; extra=='enc'"])
    repository.add_candidate("foo", "0.2.0")
    repository.add_dependencies("foo", "0.2.0", ["chardet; extra=='enc'"])
    repository.add_candidate("bar", "0.1.0")
    repository.add_dependencies("bar", "0.1.0", ["foo[enc]>=0.1.0"])
    result = resolve(["foo==0.1.0", "bar"])
    assert result["foo"].version == result["foo[enc]"].version == "0.1.0"


def test_resolve_skip_candidate_with_invalid_metadata(resolve, repository):
    repository.add_candidate("sqlparse", "0.4.0")
    repository.add_dependencies("sqlparse", "0.4.0", ["django>=1.11'"])
    result = resolve(["sqlparse"], ">=3.6")
    assert result["sqlparse"].version == "0.3.0"


def test_resolve_direct_minimal_versions(resolve, repository, project):
    repository.add_candidate("pytz", "2019.6")
    project.add_dependencies(["django"])
    result = resolve(["django"], ">=3.6", direct_minimal_versions=True)
    assert result["django"].version == "1.11.8"
    assert result["pytz"].version == "2019.6"


def test_resolve_record_markers(resolve, repository, project):
    repository.add_candidate("A", "1.0")
    repository.add_candidate("B", "1.0")
    repository.add_candidate("C", "1.0")
    repository.add_candidate("D", "1.0")
    repository.add_candidate("E", "1.0")
    repository.add_candidate("F", "1.0")
    repository.add_dependencies("A", "1.0", ["B; os_name == 'posix'", "C; os_name=='nt'"])
    # package D has transitive markers that conflict
    repository.add_dependencies("C", "1.0", ["D; os_name!='nt'", "E; python_version < '3.8'"])
    # package E has union markers
    repository.add_dependencies("B", "1.0", ["E; python_version >= '3.7'"])
    # B -> E -> F -> B has circular dependency
    repository.add_dependencies("E", "1.0", ["F; platform_machine=='x86_64'"])
    repository.add_dependencies("F", "1.0", ["B"])

    result = resolve(["A"], ">=3.6", inherit_metadata=True)
    assert result["a"].version == "1.0"
    assert "d" not in result
    assert (
        str(result["e"].req.marker)
        == 'python_version >= "3.7" and os_name == "posix" or python_version < "3.8" and os_name == "nt"'
    )
    assert (
        str(result["f"].req.marker)
        == 'python_version >= "3.7" and os_name == "posix" and platform_machine == "x86_64" or '
        'python_version < "3.8" and os_name == "nt" and platform_machine == "x86_64"'
    )
    assert (
        str(result["b"].req.marker) == 'os_name == "posix" or (os_name == "posix" or os_name == "nt") and '
        'platform_machine == "x86_64" and python_version < "3.8"'
    )


@pytest.mark.parametrize(
    "name,value,output",
    [
        ("foo", "1.0", "foo==1.0"),
        ("foo", "==1.0", "foo==1.0"),
        ("foo", ">=1.0", "foo>=1.0"),
        ("foo", "http://foobar.com", "foo @ http://foobar.com"),
    ],
)
def test_requirement_from_override(name, value, output):
    assert get_requirement_from_override(name, value) == output



import shutil

import pytest
from unearth import Link

from pdm.models.markers import EnvSpec
from pdm.models.specifiers import PySpecSet
from pdm.pytest import Distribution
from tests import FIXTURES


def test_add_package(project, working_set, dev_option, pdm):
    pdm(["add", *dev_option, "requests"], obj=project, strict=True)
    group = project.pyproject.dependency_groups["dev"] if dev_option else project.pyproject.metadata["dependencies"]

    assert group[0] == "requests>=2.19.1"
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["idna"].version == "2.7"
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


def test_add_package_no_lock(project, working_set, dev_option, pdm):
    pdm(["add", *dev_option, "--frozen-lockfile", "-v", "requests"], obj=project, strict=True)
    group = project.pyproject.dependency_groups["dev"] if dev_option else project.pyproject.metadata["dependencies"]

    assert group[0] == "requests>=2.19.1"
    assert not project.lockfile.exists()
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


def test_add_command(project, pdm, mocker):
    do_add = mocker.patch("pdm.cli.commands.add.Command.do_add")
    pdm(["add", "requests"], obj=project)
    do_add.assert_called_once()


def test_add_package_to_custom_group(project, working_set, pdm):
    pdm(["add", "requests", "--group", "test"], obj=project, strict=True)

    assert "requests" in project.pyproject.metadata["optional-dependencies"]["test"][0]
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["idna"].version == "2.7"
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


def test_add_package_to_custom_dev_group(project, working_set, pdm):
    pdm(["add", "requests", "--group", "test", "--dev"], obj=project, strict=True)

    dependencies = project.pyproject.dependency_groups["test"]
    assert "requests" in dependencies[0]
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["idna"].version == "2.7"
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package in working_set


@pytest.mark.usefixtures("vcs")
def test_add_editable_package(project, working_set, pdm):
    # Ensure that correct python version is used.
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "--dev", "demo"], obj=project, strict=True)
    pdm(["add", "-de", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project, strict=True)

    group = project.pyproject.dev_dependencies["dev"]
    assert group == ["-e git+https://github.com/test-root/demo.git#egg=demo"]
    assert not project.pyproject.dependency_groups
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["demo"].prepare(project.environment).revision == "1234567890abcdef"
    assert working_set["demo"].link_file
    assert locked_candidates["idna"].version == "2.7"
    assert "idna" in working_set

    pdm(["sync", "--no-editable"], obj=project, strict=True)
    assert not working_set["demo"].link_file


@pytest.mark.usefixtures("vcs", "working_set")
def test_add_editable_package_to_metadata_forbidden(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    result = pdm(["add", "-v", "-e", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project)
    assert "PdmUsageError" in result.stderr
    result = pdm(["add", "-v", "-Gtest", "-e", "git+https://github.com/test-root/demo.git#egg=demo"], obj=project)
    assert "PdmUsageError" in result.stderr


@pytest.mark.usefixtures("working_set", "vcs")
def test_non_editable_override_editable(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git#egg=demo"
    pdm(["add", "--dev", "-e", url], obj=project, strict=True)
    pdm(["add", "--dev", url], obj=project, strict=True)
    assert not project.get_dependencies("dev")[0].editable


@pytest.mark.usefixtures("working_set", "vcs")
def test_add_editable_normal_dev_dependency(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git#egg=demo"
    pdm(["add", "--dev", "-e", url], obj=project, strict=True)
    pdm(["add", "-d", "urllib3"], obj=project, strict=True)
    pdm(["add", "-d", "idna"], obj=project, strict=True)
    dev_group = project.pyproject.settings["dev-dependencies"]["dev"]
    pep735_group = project.pyproject.dependency_groups["dev"]
    assert dev_group == ["-e git+https://github.com/test-root/demo.git#egg=demo"]
    assert pep735_group == ["urllib3>=1.22", "idna>=2.7"]


@pytest.mark.usefixtures("working_set", "vcs")
def test_add_dev_dependency_with_existing_editables_group(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git#egg=demo"
    pdm(["add", "-dG", "editables", "-e", url], obj=project, strict=True)
    pdm(["add", "-d", "urllib3"], obj=project, strict=True)
    pdm(["add", "-dG", "named", "idna"], obj=project, strict=True)
    assert "editables" in project.pyproject.settings["dev-dependencies"]
    assert "dev" in project.pyproject.dependency_groups
    assert "named" in project.pyproject.dependency_groups
    editables_group = project.pyproject.settings["dev-dependencies"]["editables"]
    pep735_group = project.pyproject.dependency_groups["dev"]
    pep735_named_group = project.pyproject.dependency_groups["named"]
    assert editables_group == ["-e git+https://github.com/test-root/demo.git#egg=demo"]
    assert "editables" not in project.pyproject.dependency_groups
    assert pep735_group == ["urllib3>=1.22"]
    assert pep735_named_group == ["idna>=2.7"]


@pytest.mark.usefixtures("working_set")
def test_add_remote_package_url(project, dev_option, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "http://fixtures.test/artifacts/demo-0.0.1-py2.py3-none-any.whl"
    pdm(["add", *dev_option, url], obj=project, strict=True)
    group = project.pyproject.dependency_groups["dev"] if dev_option else project.pyproject.metadata["dependencies"]
    assert group[0] == f"demo @ {url}"


def test_add_no_install(project, working_set, pdm):
    pdm(["add", "--no-sync", "requests"], obj=project, strict=True)
    for package in ("requests", "idna", "chardet", "urllib3", "certifi"):
        assert package not in working_set


def test_add_no_sync_frozen_lockfile_skips_resolution(project, pdm, mocker):
    do_lock = mocker.patch("pdm.cli.actions.do_lock")
    pdm(["add", "--frozen-lockfile", "--no-sync", "requests"], obj=project, strict=True)

    do_lock.assert_not_called()
    assert project.pyproject.metadata["dependencies"][0] == "requests"
    assert not project.lockfile.exists()


@pytest.mark.usefixtures("repository")
def test_add_package_save_exact(project, pdm):
    pdm(["add", "--save-exact", "--no-sync", "requests"], obj=project, strict=True)
    assert project.pyproject.metadata["dependencies"][0] == "requests==2.19.1"


@pytest.mark.usefixtures("repository")
def test_add_package_save_wildcard(project, pdm):
    pdm(["add", "--save-wildcard", "--no-sync", "requests"], obj=project, strict=True)
    assert project.pyproject.metadata["dependencies"][0] == "requests"


@pytest.mark.usefixtures("repository")
def test_add_package_save_minimum(project, pdm):
    pdm(["add", "--save-minimum", "--no-sync", "requests"], obj=project, strict=True)
    assert project.pyproject.metadata["dependencies"][0] == "requests>=2.19.1"


def test_add_package_update_reuse(project, repository, pdm):
    pdm(["add", "--no-sync", "--save-wildcard", "requests", "pytz"], obj=project, strict=True)

    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.19.1"
    assert locked_candidates["chardet"].version == "3.0.4"
    assert locked_candidates["pytz"].version == "2019.3"

    repository.add_candidate("pytz", "2019.6")
    repository.add_candidate("chardet", "3.0.5")
    repository.add_candidate("requests", "2.20.0")
    repository.add_dependencies(
        "requests",
        "2.20.0",
        [
            "certifi>=2017.4.17",
            "chardet<3.1.0,>=3.0.2",
            "idna<2.8,>=2.5",
            "urllib3<1.24,>=1.21.1",
        ],
    )
    pdm(["add", "--no-sync", "--save-wildcard", "requests"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.20.0"
    assert locked_candidates["chardet"].version == "3.0.4"
    assert locked_candidates["pytz"].version == "2019.3"


def test_add_package_update_eager(project, repository, pdm):
    pdm(["add", "--no-sync", "--save-wildcard", "requests", "pytz"], obj=project, strict=True)

    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.19.1"
    assert locked_candidates["chardet"].version == "3.0.4"
    assert locked_candidates["pytz"].version == "2019.3"

    repository.add_candidate("pytz", "2019.6")
    repository.add_candidate("chardet", "3.0.5")
    repository.add_candidate("requests", "2.20.0")
    repository.add_dependencies(
        "requests",
        "2.20.0",
        [
            "certifi>=2017.4.17",
            "chardet<3.1.0,>=3.0.2",
            "idna<2.8,>=2.5",
            "urllib3<1.24,>=1.21.1",
        ],
    )
    pdm(["add", "--no-sync", "--save-wildcard", "--update-eager", "requests"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["requests"].version == "2.20.0"
    assert locked_candidates["chardet"].version == "3.0.5"
    assert locked_candidates["pytz"].version == "2019.3"


def test_add_package_with_mismatch_marker(project, working_set, pdm):
    env = project.environment
    env.__dict__["spec"] = EnvSpec.from_spec("==3.11", "macos", "cpython")
    pdm(["add", "requests", "pytz; platform_system!='Darwin'"], obj=project, strict=True)
    assert "pytz" not in working_set


def test_add_dependency_from_multiple_parents(project, working_set, pdm):
    env = project.environment
    env.__dict__["spec"] = EnvSpec.from_spec("==3.11", "macos", "cpython")
    pdm(["add", "requests", "chardet; platform_system!='Darwin'"], obj=project, strict=True)
    assert "chardet" in working_set


def test_add_packages_without_self(project, working_set, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "--no-self", "requests"], obj=project, strict=True)
    assert project.name not in working_set


@pytest.mark.usefixtures("working_set")
def test_add_package_unconstrained_rewrite_specifier(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "--no-self", "django"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["django"].version == "2.2.9"
    assert project.pyproject.metadata["dependencies"][0] == "django>=2.2.9"

    pdm(["add", "--no-self", "--unconstrained", "django-toolbar"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["django"].version == "1.11.8"
    assert project.pyproject.metadata["dependencies"][0] == "django>=1.11.8"


@pytest.mark.usefixtures("working_set", "vcs")
def test_add_cached_vcs_requirement(project, mocker, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    url = "git+https://github.com/test-root/demo.git@1234567890abcdef#egg=demo"
    built_path = FIXTURES / "artifacts/demo-0.0.1-py2.py3-none-any.whl"
    wheel_cache = project.make_wheel_cache()
    cache_path = wheel_cache.get_path_for_link(Link(url), project.environment.spec)
    if not cache_path.exists():
        cache_path.mkdir(parents=True)
    shutil.copy2(built_path, cache_path)
    downloader = mocker.patch("unearth.finder.unpack_link")
    builder = mocker.patch("pdm.builders.WheelBuilder.build")
    pdm(["add", "--no-self", url], obj=project, strict=True)
    lockfile_entry = next(p for p in project.lockfile["package"] if p["name"] == "demo")
    assert lockfile_entry["revision"] == "1234567890abcdef"
    downloader.assert_not_called()
    builder.assert_not_called()


@pytest.mark.usefixtures("repository")
def test_add_with_dry_run(project, pdm):
    result = pdm(["add", "--dry-run", "requests"], obj=project, strict=True)
    project.pyproject.reload()
    assert not project.get_dependencies()
    assert "requests 2.19.1" in result.stdout
    assert "urllib3 1.22" in result.stdout


def test_add_with_prerelease(project, working_set, pdm):
    pdm(["add", "--prerelease", "--save-compatible", "urllib3"], obj=project, strict=True)
    assert working_set["urllib3"].version == "1.23b0"
    assert project.pyproject.metadata["dependencies"][0] == "urllib3<2,>=1.23b0"


def test_add_editable_package_with_extras(project, working_set, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    dep_path = FIXTURES.joinpath("projects/demo")
    pdm(["add", "-dGdev", "-e", f"{dep_path.as_posix()}[security]"], obj=project, strict=True)
    assert f"-e {dep_path.as_uri()}#egg=demo[security]" in project.use_pyproject_dependencies("dev", True)[0]
    assert "demo" in working_set
    assert "requests" in working_set
    assert "urllib3" in working_set


@pytest.mark.usefixtures("repository")
def test_add_dependency_with_extras(project, pdm):
    project.environment.python_requires = PySpecSet(">=3.6")
    pdm(["add", "requests[security]", "--no-sync"], obj=project, strict=True)
    locked_repo = project.get_locked_repository()
    extra_package = next(p for p in locked_repo.packages.values() if p.candidate.identify() == "requests[security]")
    assert extra_package.dependencies == ["pyOpenSSL>=0.14", "requests==2.19.1"]
    # test adding new package won't add duplicate dependencies
    pdm(["add", "pytz", "--no-sync"], obj=project, strict=True)
    locked_repo = project.get_locked_repository()
    extra_package = next(p for p in locked_repo.packages.values() if p.candidate.identify() == "requests[security]")
    assert extra_package.dependencies == ["pyOpenSSL>=0.14", "requests==2.19.1"]


def test_add_package_with_local_version(project, repository, working_set, pdm):
    repository.add_candidate("foo", "1.0a0+local")
    pdm(["add", "-v", "foo"], obj=project, strict=True)
    assert working_set["foo"].version == "1.0a0+local"
    dependencies, _ = project.use_pyproject_dependencies("default")
    assert dependencies[0] == "foo>=1.0a0"


def test_add_group_to_lockfile(project, working_set, pdm):
    pdm(["add", "requests"], obj=project, strict=True)
    assert project.lockfile.groups == ["default"]
    pdm(["add", "--group", "tz", "pytz"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "tz"]
    assert "pytz" in working_set


def test_add_group_to_lockfile_without_package(project, working_set, pdm):
    project.add_dependencies(["requests"])
    project.add_dependencies(["pytz"], to_group="tz")
    pdm(["install"], obj=project, strict=True)
    assert "pytz" not in working_set
    assert project.lockfile.groups == ["default"]
    pdm(["add", "--group", "tz"], obj=project, strict=True)
    assert project.lockfile.groups == ["default", "tz"]
    assert "pytz" in working_set


def test_add_update_reuse_installed(project, working_set, repository, pdm):
    working_set["foo"] = Distribution("foo", "1.0.0")
    repository.add_candidate("foo", "1.0.0")
    repository.add_candidate("foo", "1.1.0")
    pdm(["add", "--update-reuse-installed", "foo"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["foo"].version == "1.0.0"


def test_add_update_reuse_installed_config(project, working_set, repository, pdm):
    working_set["foo"] = Distribution("foo", "1.0.0")
    repository.add_candidate("foo", "1.0.0")
    repository.add_candidate("foo", "1.1.0")
    project.project_config["strategy.update"] = "reuse-installed"
    pdm(["add", "foo"], obj=project, strict=True)
    locked_candidates = project.get_locked_repository().candidates
    assert locked_candidates["foo"].version == "1.0.0"


def test_add_disable_cache(project, pdm, working_set):
    cache_dir = project.cache_dir
    pdm(["--no-cache", "add", "requests"], obj=project, strict=True)
    assert "requests" in working_set

    files = [file for file in cache_dir.rglob("*") if file.is_file()]
    assert not files


@pytest.mark.usefixtures("working_set")
def test_add_dependency_with_direct_minimal_versions(project, pdm, repository):
    pdm(["lock", "-S", "direct_minimal_versions"], obj=project, strict=True)
    repository.add_candidate("pytz", "2019.6")
    pdm(["add", "django"], obj=project, strict=True)
    all_candidates = project.get_locked_repository().candidates
    assert "django>=1.11.8" in project.pyproject.metadata["dependencies"]
    assert all_candidates["django"].version == "1.11.8"
    assert all_candidates["pytz"].version == "2019.6"


def test_add_group_with_normalized_name(project, pdm, working_set):
    project.pyproject.dependency_groups.update({"foo_bar": ["requests"]})
    project.pyproject.write()
    pdm(["lock"], obj=project, strict=True)
    assert "foo-bar" in project.lockfile.groups
    pdm(["sync", "-G", "foo.bar"], obj=project, strict=True)
    assert "requests" in working_set
    result = pdm(["add", "-G", "foo-bar", "pytz"], obj=project)
    assert result.exit_code != 0
    assert "Group foo-bar already exists in another non-normalized form" in result.stderr


@pytest.mark.usefixtures("working_set")
def test_add_to_dependency_group_with_include(project, pdm):
    from pdm.formats.base import make_array

    project.pyproject.dependency_groups.update({"tz": ["pytz"], "web": make_array([{"include-group": "tz"}])})
    project.pyproject.write()
    pdm(["add", "-Gweb", "requests"], obj=project, strict=True)
    assert project.pyproject.dependency_groups["web"] == [{"include-group": "tz"}, "requests>=2.19.1"]



from setuptools import setup


setup(
    name="demo-extras",
    version="0.0.1",
    description="test demo",
    py_modules=["demo"],
    install_requires=[],
    extras_require={"extra1": ["requests[security]"], "extra2": ["requests[socks]"]},
)



"""Tests for the completion command"""


def test_completion_bash(pdm):
    """Test completion for bash shell"""
    result = pdm(["completion", "bash"])
    assert result.exit_code == 0
    assert "BASH completion script for pdm" in result.output


def test_completion_zsh(pdm):
    """Test completion for zsh shell"""
    result = pdm(["completion", "zsh"])
    assert result.exit_code == 0
    assert "#compdef pdm" in result.output


def test_completion_fish(pdm):
    """Test completion for fish shell"""
    result = pdm(["completion", "fish"])
    assert result.exit_code == 0
    assert "FISH completion script for pdm" in result.output


def test_completion_powershell(pdm):
    """Test completion for powershell"""
    result = pdm(["completion", "powershell"])
    assert result.exit_code == 0
    assert "Powershell completion script for pdm" in result.output


def test_completion_pwsh(pdm):
    """Test completion for pwsh (PowerShell Core)"""
    result = pdm(["completion", "pwsh"])
    assert result.exit_code == 0
    assert "Powershell completion script for pdm" in result.output


def test_completion_unsupported_shell(pdm):
    """Test completion with unsupported shell raises error"""
    result = pdm(["completion", "unsupported_shell"])
    assert result.exit_code != 0
    assert "Unsupported shell" in result.stderr


def test_completion_auto_detect(pdm, monkeypatch):
    """Test completion with auto-detected shell"""
    import shellingham

    monkeypatch.setattr(shellingham, "detect_shell", lambda: ("bash", "/bin/bash"))
    result = pdm(["completion"])
    assert result.exit_code == 0
    assert "BASH completion script for pdm" in result.output


def test_completion_auto_detect_unsupported(pdm, monkeypatch):
    """Test completion with auto-detected unsupported shell"""
    import shellingham

    monkeypatch.setattr(shellingham, "detect_shell", lambda: ("csh", "/bin/csh"))
    result = pdm(["completion"])
    assert result.exit_code != 0
    assert "Unsupported shell" in result.stderr



import pytest

from pdm.models.markers import EnvSpec, get_marker
from pdm.models.specifiers import PySpecSet


@pytest.mark.parametrize(
    "original,marker,py_spec",
    [
        ("python_version > '3'", "", ">=3.1"),
        ("python_version > '3.8'", "", ">=3.9"),
        ("python_version != '3.8'", "", "!=3.8.*"),
        ("python_version == '3.7'", "", "==3.7.*"),
        ("python_version in '3.6 3.7'", "", ">=3.6.0,<3.8.0"),
        ("python_full_version >= '3.6.0'", "", ">=3.6"),
        ("python_full_version not in '3.8.3'", "", "!=3.8.3"),
        # mixed marker and python version
        ("python_version > '3.7' and os_name == 'nt'", 'os_name == "nt"', ">=3.8"),
        (
            "python_version > '3.7' or os_name == 'nt'",
            'python_version > "3.7" or os_name == "nt"',
            "",
        ),
    ],
)
def test_split_pyspec(original, marker, py_spec):
    m = get_marker(original)
    a, b = m.split_pyspec()
    assert marker == str(a)
    assert b == PySpecSet(py_spec)


@pytest.mark.parametrize(
    "marker,env_spec,expected",
    [
        ("os_name == 'nt'", EnvSpec.from_spec(">=3.10", "windows"), True),
        ("os_name == 'nt'", EnvSpec.from_spec(">=3.10"), True),
        ("os_name != 'nt'", EnvSpec.from_spec(">=3.10", "windows"), False),
        ("python_version >= '3.7' and os_name == 'nt'", EnvSpec.from_spec(">=3.10"), True),
        ("python_version < '3.7' and os_name == 'nt'", EnvSpec.from_spec(">=3.10"), False),
        ("python_version < '3.7' or os_name == 'nt'", EnvSpec.from_spec(">=3.10"), False),
        ("python_version >= '3.7' and os_name == 'nt'", EnvSpec.from_spec(">=3.10", "linux"), False),
        ("python_version >= '3.7' or os_name == 'nt'", EnvSpec.from_spec(">=3.10", "linux"), True),
        ("python_version >= '3.7' and implementation_name == 'pypy'", EnvSpec.from_spec(">=3.10"), True),
        (
            "python_version >= '3.7' and implementation_name == 'pypy'",
            EnvSpec.from_spec(">=3.10", implementation="cpython"),
            False,
        ),
    ],
)
def test_match_env_spec(marker, env_spec, expected):
    m = get_marker(marker)
    assert m.matches(env_spec) is expected



from setuptools import setup

import first

setup(
    name="demo",
    version="0.0.1",
    description="test demo",
    py_modules=["demo"],
    python_requires=">=3.3",
    install_requires=["idna", "chardet; os_name=='nt'"],
    extras_require={
        "tests": ["pytest"],
        "security": ['requests; python_version>="3.6"'],
    },
)
